"""Classes for representing guide RNA libraries."""

from typing import Dict, Iterable, Optional, Tuple, Union

from dataclasses import asdict, dataclass, field
import math

from bioino import FastaSequence, GffFile, GffLine
from carabiner import pprint_dict
from nemony import encode, hash as nm_hash
from streq import find_iupac, reverse_complement, which_re_sites
from tqdm.auto import tqdm

from .annotate import annotate_from_gff
from .features import get_context

@dataclass
class GuideMatch:

    """Information of guide matching a genome.

    Attributes
    ----------
    pam_search : str
        IUPAC search string for PAM.
    guide_seq : str
        Guide spacer sequence.
    pam_seq : str
        Actual PAM sequence.
    pam_start : int
        Chromosome coordinate of PAM start.
    pam_end : int
        Chromosome coordinate of PAM end.
    length : int
        Length of guide.

    Examples
    --------
    >>> GuideMatch(pam_search='NGG', guide_seq='ATCGATCG', pam_seq='CGG', pam_start=10, reverse=False)
    GuideMatch(pam_search='NGG', guide_seq='ATCGATCG', pam_seq='CGG', pam_start=10, reverse=False, guide_context_up=None, guide_context_down=None, pam_end=13, length=8, guide_start=2, guide_end=10)
    >>> GuideMatch(pam_search='NGG', guide_seq='ATCGATCG', pam_seq='CCG', pam_start=10, reverse=True)
    GuideMatch(pam_search='NGG', guide_seq='CGATCGAT', pam_seq='CGG', pam_start=10, reverse=True, guide_context_up=None, guide_context_down=None, pam_end=13, length=8, guide_start=13, guide_end=21)
    
    """

    pam_search: str
    guide_seq: str
    pam_seq: str 
    pam_start: int
    reverse: bool
    guide_context_up: Optional[str] = field(init=False, default=None)
    guide_context_down: Optional[str] = field(init=False, default=None)
    pam_end: int = field(init=False)
    length: int = field(init=False)
    guide_start: int = field(init=False)
    guide_end: int = field(init=False)

    def __post_init__(self):

        self.length = len(self.guide_seq)
        self.pam_end = self.pam_start + len(self.pam_seq)

        if not self.reverse:
            self.guide_start = self.pam_start - self.length 
            self.guide_end = self.pam_start 
        else:
            self.guide_start = self.pam_end
            self.guide_end = self.pam_end + self.length
            self.pam_seq = reverse_complement(self.pam_seq)
            self.guide_seq = reverse_complement(self.guide_seq)

    def __len__(self):

        return len(self.guide_seq)

    def __str__(self):

        return self.guide_seq
        
    def as_dict(self):

        return asdict(self)


@dataclass
class GuideMatchCollection:

    """Set of guides with the same sequence but potentially with multiple 
    matches.

    Attributes
    ----------
    guide_seq : str
        Guide spacer sequence.
    pam_search : str
        IUPAC search string for PAM.
    matches : iterable of GuideMatch
        Objects with matching information.
    guide_name : str, optional
        Name or identifier of guide.

    """

    guide_seq: str
    pam_search: str
    matches: Iterable[GuideMatch]
    guide_name: Optional[str] = field(default=None)

    def __iter__(self):

        try:
            for match in self.matches:
                yield match
        except ValueError as e:
            raise e

    def __len__(self):

        """Give the number of matches.
        
        If the matches are not yet instantiated, this will instantiate the matches. 
        Depending on the size of the search, it can take memory and time.
        
        """

        try:
            return len(self.matches)
        except TypeError:
            self.matches = list(self.matches)
            return len(self.matches)            

    @staticmethod
    def _from_search(guide_seq: str, 
                     genome: str,
                     pam_search: str = "NGG") -> Iterable[Tuple[int, Dict[str, GuideMatch]]]:
    
        if (guide_seq not in genome and 
            reverse_complement(guide_seq) not in genome):
            raise ValueError(f'{guide_seq} not in genome')

        pam_len = len(pam_search)

        guide_plus_pam = guide_seq + pam_search

        for reverse in (False, True):

            this_guide_plus_pam = (reverse_complement(guide_plus_pam) if reverse 
                                   else guide_plus_pam)
            
            for i, ((guide_pam_start, guide_pam_end), 
                     guide_pam_seq) in enumerate(find_iupac(this_guide_plus_pam, genome)):
                
                if not reverse:
                    pam_start = guide_pam_end - pam_len
                    pam_seq = guide_pam_seq[-pam_len:]
                    _guide_seq = guide_pam_seq[:-pam_len]
                else:
                    pam_start = guide_pam_start
                    pam_seq = guide_pam_seq[:pam_len]
                    _guide_seq = guide_pam_seq[pam_len:]

                gm = GuideMatch(pam_search=pam_search,
                                pam_seq=pam_seq, 
                                guide_seq=_guide_seq,
                                pam_start=pam_start, 
                                reverse=reverse)

                guide_down, guide_up = get_context(gm.pam_start, gm.pam_end,
                                                   gm.guide_start, gm.guide_end,
                                                   genome=genome, 
                                                   reverse=reverse)
                gm.guide_context_down = guide_down
                gm.guide_context_up = guide_up

                yield gm

        return i

    @classmethod
    def from_search(cls,
                    guide_seq: str,
                    genome: str,
                    pam_search: str = "NGG",
                    guide_name: Optional[str] = None,
                    in_memory: bool = False):
        
        """Find the location of a guide sequence in a genome.

        Searches the genome in the forward strand then the reverse strand,
        returning the match with an adjacent PAM in the order found.

        The default behavior is to find matches lazily to save memory and time.

        Parameters
        ----------
        guide_seq : str
            The sequence of the guide to be found.
        pam_search : str, optional
            The sequence (IUPAC codes allowed) of the PAM to match. Default: "NGG".
        genome : str
            The genome sequence to search.
        guide_name : str
            Name or identifier of guide.

        Raises
        ------
        ValueError
            If guide not found in genome with appropriate PAM.

        Returns
        -------
        GuideMatches
            A iterator of dictionaries of match information.

        Examples
        --------
        >>> gmc = GuideMatchCollection.from_search("TTTTTTTAAAAAAA", "CCGTTTTTTTAAAAAAACGG")
        >>> len(gmc)
        2
        >>> for match in gmc:
        ...     print(match)
        ... 
        TTTTTTTAAAAAAA
        TTTTTTTAAAAAAA

        """

        matches = cls._from_search(guide_seq, genome, pam_search)

        if in_memory:
            matches = list(matches)

        return cls(pam_search=pam_search, 
                   guide_seq=guide_seq, 
                   guide_name=guide_name,
                   matches=matches)
    

@dataclass
class GuideLibrary:

    """Library of guides from a genome.

    Attributes
    ----------
    genome : str
        Genome sequence that guides are matched to.
    guide_matches : list of GuideMatchCollection
        List of matches to the genome.

    """

    genome: str
    guide_matches: Iterable[GuideMatchCollection]

    def __iter__(self):
        
        try:
            for match in self.guide_matches:
                yield match
        except ValueError as e:
            raise e
    
    def __len__(self):

        """Give the number of matches.
        
        If the matches are not yet instantiated, this will instantiate the matches. 
        Depending on the size of the search, it can take memory and time.
        
        """

        try:
            return len(self.guide_matches)
        except TypeError:
            self.guide_matches = list(self.guide_matches)
            return len(self.guide_matches)

    def as_gff(self, 
               max: Optional[int] = None,
               annotations_from: Optional[GffFile] = None,
               tags: Optional[Iterable[str]] = None,
               gff_defaults: Optional[Dict[str, Union[str, int]]] = None) -> Iterable[GffLine]:

        """Convert into a iterable of `bioino.GffLine`s.

        Parameters
        ----------
        max : int, optional
            Number of `bioino.GffLine`s to return for each `GuideMatchCollection`. Default: return all.
        annotations_from : bioino.GffFile, optional
            If provided use the `lookup` table to annotate the returned `GffLine`s.
        tags : list of str, optional
            Which tags to take from `annotations_from`.
        gff_defaults : dict
            In case of missing values that are essential for GFF file formats
            (namely columns 1-8), take values from this disctionary.

        Yields
        ------
        bioino.GffLine
            Corresponding to a `GuideMatch`.

        Examples
        --------
        >>> genome = "ATATATATATATATATATATATATACCGTTTTTTTAAAAAAACGGATATATATATATAATATATATATATAATATATATATATA"
        >>> gl = GuideLibrary.from_generating(genome=genome)
        >>> for gff in gl.as_gff(gff_defaults=dict(seqid='my_seq', source='here', feature='protospacer')):  # doctest: +NORMALIZE_WHITESPACE
        ...     print(gff)
        ... 
        my_seq    here    protospacer     23      42      .       +       .       ID=sgr-06a4ba9b;Name=42-united_exodus;guide_context_down=ATATATATATATAATATATA;guide_context_up=ATATATATATATATATATAT;guide_length=20;guide_re_sites=;guide_sequence=ATACCGTTTTTTTAAAAAAA;guide_sequence_hash=a3987295;mnemonic=united_exodus;pam_end=45;pam_replichore=L;pam_search=NGG;pam_sequence=CGG;pam_start=42;source_name=42-united_exodus
        my_seq    here    protospacer     29      48      .       -       .       ID=sgr-f84d1c6a;Name=25-zigzag_state;guide_context_down=TATATATATATATATATATA;guide_context_up=ATATATATATTATATATATA;guide_length=20;guide_re_sites=;guide_sequence=TATCCGTTTTTTTAAAAAAA;guide_sequence_hash=188c9ee6;mnemonic=zigzag_state;pam_end=28;pam_replichore=R;pam_search=NGG;pam_sequence=CGG;pam_start=25;source_name=25-zigzag_state       
        """

        genome_length = len(self.genome)
        max = max or math.inf
        gff_defaults = gff_defaults or {}

        for guide_match_collection in self:

            for i, match in enumerate(guide_match_collection.matches):

                if i >= max:
                    break

                sgrna_info = {
                    'ID': 'sgr-' + nm_hash((match.guide_seq, match.pam_search, match.pam_start), 8),
                    'mnemonic': encode((match.guide_seq, match.pam_search, match.pam_start)),
                    'guide_sequence_hash': nm_hash(match.guide_seq, 8),
                    'source_name': guide_match_collection.guide_name,
                    'pam_start': match.pam_start, 
                    'pam_end': match.pam_end,
                    'pam_search': match.pam_search, 
                    'pam_sequence': match.pam_seq,
                    'pam_replichore': 'R' if ((match.pam_start / genome_length) < 0.5) else 'L',
                    'strand': ('+' if not match.reverse else '-'),
                    'start': ((match.guide_start + 1) if (match.guide_start + 1) > 0 
                            else match.guide_start + 1 + genome_length), 
                    'end': (match.guide_end if (match.guide_start + 1) > 0 
                            else match.guide_end + genome_length), 
                    'guide_context_up': match.guide_context_up, 
                    'guide_context_down': match.guide_context_down,
                    'guide_length': match.length,
                    'guide_re_sites': ','.join(which_re_sites(match.guide_seq)),
                    'guide_sequence': match.guide_seq
                }
                sgrna_info['Name'] = '{pam_start}-{mnemonic}'.format(**sgrna_info)

                if annotations_from is not None:

                    sgrna_info = annotate_from_gff(sgrna_info, 
                                                   gff_data=annotations_from, 
                                                   tags=tags)
                    sgrna_info['Name'] = '{ann_Name}-{pam_start}-{mnemonic}'.format(**sgrna_info)
                
                sgrna_info.update(gff_defaults)
                sgrna_info['source_name'] = sgrna_info['source_name'] or sgrna_info['Name']

                yield GffLine.from_dict(sgrna_info)

    @staticmethod
    def _from_mapping(guide_seq: Iterable[FastaSequence],
                      genome: str,
                      pam_search: str = "NGG"):
        
        not_found = {}

        with tqdm(guide_seq) as t:  ## run a progress bar
    
            for guide_sequence in t:
                
                t.set_postfix(current=guide_sequence.name[:40], 
                              not_found=len(not_found))

                try:
                    guide_matches = GuideMatchCollection.from_search(guide_seq=guide_sequence.sequence, 
                                                                     guide_name=guide_sequence.name,
                                                                     pam_search=pam_search, 
                                                                     genome=genome, 
                                                                     in_memory=True)  ## forces the ValueError to be raised now
                except ValueError:
                    not_found[guide_sequence.name] = guide_sequence.sequence
                else:
                    yield guide_matches

        pprint_dict(not_found, 
                    message=f'Not found: {len(not_found)} guides')

        return not_found

    @classmethod
    def from_mapping(cls,
                     guide_seq: Union[str, Iterable[str], FastaSequence, Iterable[FastaSequence]],
                     genome: str,
                     pam_search: str = "NGG",
                     in_memory: bool = False):
        
        """Map a set of expected guides to a genome.

        The default behavior is to find matches lazily to save memory and time.
        
        Parameters
        ----------
        guide_seq : str or bioino.FastaSequence or list
            Guides to map.
        genome : str
            Genome to map against.
        pam_search : str
            IUPAC PAM sequence to search against.
        in_memory : bool, optional
            Whether to instantiate matches in memory. Default: lazy matching.

        Returns
        -------
        GuideLibrary

        Examples
        --------
        >>> genome = "CCCCCCCCCCCTTTTTTTTTTAAAAAAAAAATGATCGATCGATCGAGGAAAAAAAAAACCCCCCCCCCC"
        >>> guide_seq = ["ATGATCGATCGATCG", "ATGATCGATCGATCGCCC"]
        >>> gl = GuideLibrary.from_mapping(guide_seq=guide_seq, genome=genome) 
        >>> for collection in gl:
        ...     for match in collection:
        ...             print(match.as_dict())
        ...
        {'pam_search': 'NGG', 'guide_seq': 'ATGATCGATCGATCG', 'pam_seq': 'AGG', 'pam_start': 45, 'reverse': False, 'guide_context_up': 'CTTTTTTTTTTAAAAAAAAA', 'guide_context_down': 'AAAAAAAAAACCCCCCCCCC', 'pam_end': 48, 'length': 15, 'guide_start': 30, 'guide_end': 45}

        """
        
        if isinstance(guide_seq, str):

            guide_seq = [guide_seq]
            
        if isinstance(guide_seq, Iterable):
            
            new_guide_seq = []

            for g in guide_seq:

                if isinstance(g, str):

                    g = FastaSequence(name=g, 
                                      description='query_spacer', 
                                      sequence=g)
                
                new_guide_seq.append(g)

            guide_seq = new_guide_seq
                
        matches = cls._from_mapping(guide_seq, genome, pam_search)

        if in_memory:
            matches = list(matches)
        
        return cls(genome=genome,
                   guide_matches=matches)
    
    @staticmethod
    def _from_generating(genome: str,
                         max_length: int = 20,
                         min_length: Optional[int] = None, 
                         pam_search: str = "NGG") -> Iterable[GuideMatchCollection]:

        min_length = min_length or max_length
        found, guides_created = 0, 0
        
        for reverse in (False, True):

            directionality = 'reverse' if reverse else 'forward'
        
            _pam_search = (reverse_complement(pam_search) if reverse 
                           else pam_search)

            with tqdm(find_iupac(_pam_search, genome)) as t:  ## run a progress bar
                
                for (pam_start, pam_end), pam_seq in t:

                    found += 1

                    for length in range(min_length, max_length + 1):

                        guides_created += 1
                        
                        guide_start = (pam_start - length if not reverse 
                                       else pam_end)
                        guide_end = (pam_start if not reverse else 
                                     pam_end + length)

                        guide_seq = genome[guide_start:guide_end]
                        
                        t.set_postfix(direction=directionality,
                                      at_site=pam_start,
                                      pam_sites_found=found,
                                      guides_created=guides_created)
                        
                        gm = GuideMatch(pam_search=pam_search,
                                                 pam_seq=pam_seq, 
                                                 guide_seq=guide_seq,
                                                 pam_start=pam_start, 
                                                 reverse=reverse)
                        guide_down, guide_up = get_context(gm.pam_start, gm.pam_end,
                                                           gm.guide_start, gm.guide_end,
                                                           genome=genome, 
                                                           reverse=reverse)
                        gm.guide_context_down = guide_down
                        gm.guide_context_up = guide_up

                        # TODO: Actually group by sequence
                        yield GuideMatchCollection(guide_seq=guide_seq, 
                                                   pam_search=_pam_search,
                                                   matches=[gm])
                        
        return guides_created
                        
    @classmethod
    def from_generating(cls,
                        genome: str,
                        max_length: int = 20,
                        min_length: Optional[int] = None, 
                        pam_search: str = "NGG",
                        in_memory: bool = False):
        
        """Find all guides matching a PAM sequence in a given genome.

        The default behavior is to find matches lazily to save memory and time.

        Parameters
        ----------
        genome : str
            Genome sequence to search.
        max_length : int, optional
            Maximum guide length. Default: 20.
        min_length : int, optional
            Minimum guide length. Default: same as max_length.
        pam_search : str, optional
            IUPAC PAM sequence to search for. Default: "NGG".
        in_memory : bool, optional
            Whether to instantiate matches in memory. Default: lazy matching.

        Examples
        --------
        >>> genome = "ATATATATATATATATATATATATACCGTTTTTTTAAAAAAACGGATATATATATATAATATATATATATAATATATATATATA"
        >>> gl = GuideLibrary.from_generating(genome=genome)
        >>> len(gl)
        2
        >>> for match_collection in gl:
        ...     for guide in match_collection:
        ...             print(guide)
        ... 
        ATACCGTTTTTTTAAAAAAA
        TATCCGTTTTTTTAAAAAAA

        """
        
        matches = (match for match in cls._from_generating(genome, max_length, min_length, pam_search))

        if in_memory:
            matches = list(matches)

        return cls(genome=genome, guide_matches=matches)



from typing import (Any, Callable, Dict, Iterable, List, 
                    Tuple, Optional, Union)

from itertools import dropwhile, takewhile

from bioino import GffLine
from streq import (correlation, gc_content, purine_content, 
                   reverse_complement)

_FLOAT_FORMAT = '.3f'

def _format_float(f: Callable[[Any], float]) -> Callable[[Any], float]:

    def _f(*args, **kwargs):

        return format(f(*args, **kwargs), _FLOAT_FORMAT)

    return _f


@_format_float
def _autocorrelation(gff: GffLine, 
                     attribute: str,
                     scaffold: Optional[str] = None) -> float:
    
    seq = gff.attributes[attribute]

    return correlation(seq, 
                       scaffold or reverse_complement(seq),
                       wobble=True)


def _pam_leading_n(gff: GffLine,
                   scaffold: Optional[str] = None) -> str:

    x = takewhile(lambda x: x[0] == 'N', 
                  zip(gff.attributes['pam_search'], 
                      gff.attributes['pam_sequence']))

    return ''.join(item[1] for item in x)


def _pam_defined(gff: GffLine, 
                 scaffold: Optional[str] = None) -> str:

    x = dropwhile(lambda x: x[0] == 'N', 
                  zip(gff.attributes['pam_search'], 
                      gff.attributes['pam_sequence']))

    return ''.join(item[1] for item in x)


_FEATURIZERS = {'on_nontemplate_strand': lambda gff, _: gff.columns.strand != gff.attributes['ann_strand'],
                'context_up2': lambda gff, _: gff.attributes['guide_context_up'][-2:],
                'context_down2': lambda gff, _: gff.attributes['guide_context_down'][:2],
                'context_up_autocorr': lambda gff, _: _autocorrelation(gff, 'guide_context_up'),
                'pam_n': _pam_leading_n,
                'pam_def': _pam_defined,
                'pam_gc': lambda gff, _: _format_float(gc_content)(gff.attributes['pam_sequence']),
                'pam_autocorr': lambda gff, _: _autocorrelation(gff, 'pam_sequence'),
                'pam_scaff_corr': lambda gff, scaffold: _autocorrelation(gff, 'pam_sequence', scaffold),
                'guide_purine': lambda gff, _: _format_float(purine_content)(gff.attributes['guide_sequence']),
                'guide_gc': lambda gff, _: _format_float(gc_content)(gff.attributes['guide_sequence']),
                'seed_seq': lambda gff, _: gff.attributes['guide_sequence'][-5:],
                'guide_start3': lambda gff, _: gff.attributes['guide_sequence'][:3],
                'guide_end3': lambda gff, _: gff.attributes['guide_sequence'][-3:],
                'guide_autocorr': lambda gff, _: _autocorrelation(gff, 'guide_sequence'),
                'guide_scaff_corr': lambda gff, scaffold: _autocorrelation(gff, 'guide_sequence', scaffold)}


def get_features() -> List[str]:

    """Get the list of available features."""

    return list(_FEATURIZERS)


def featurize(gff: GffLine, 
              features: Optional[Union[str, Iterable[str]]] = None,
              scaffold: Optional[str] = None) -> Union[int, str, Dict[str, Union[int, str]]]:
    
    """Featurize a guide RNA represented by a `bioino.GffLine`.

    Depending on the feature to be calculated, the GFF should have attributes
    'pam_sequence', 'guide_sequence', 'guide_context_up', 'guide_context_down',
    and 'ann_strand'.

    Parameters
    ----------
    gff : bioino.GffLine
        Input guide RNA with additional attributes.
    features : str or list of str, optional
        The names of the features to be calculated. Default: calculate all.
    scaffold : str, optional
        Guide scaffold. Required for some features. If `features` is the default,
        scaffold must be provided.

    Returns
    -------
    dict, float, or str
        If `features` is a string, then returns the value of the feature.
        If it is a list, then returns a dictionary mapping feature names
        to values.

    Raises
    ------
    KeyError
        If any `features` are not supported.
    ValueError
        If `features` is neither a string nor iterable.
    AttributeError
        If `features` is default but `scaffold` is not provided.
    
    """
    
    if features is None:

        features = features or get_features()

        if scaffold is None:

            raise AttributeError("Scaffold must be provided to calculate all features.")

    if isinstance(features, str):

        return _FEATURIZERS[features](gff, scaffold)
    
    elif isinstance(features, Iterable):

        return {f"feat_{feature}": _FEATURIZERS[feature](gff, scaffold)
                for feature in features}
    
    else:

        raise ValueError(f"Requested feature {features} is not a string or iterable.")


def get_context(pam_start: int, 
                pam_end: int,
                guide_start: int, 
                guide_end: int,
                genome: str,
                reverse: bool,
                extra_bases: int = 20) -> Tuple[str, str]:
    
    if not reverse:
        guide_down = genome[pam_end:(pam_end + extra_bases)] 
        guide_up = genome[(guide_start - extra_bases):guide_start]
    else:
        guide_down = reverse_complement(genome[(pam_start - extra_bases):pam_start])
        guide_up = reverse_complement(genome[guide_end:(guide_end + extra_bases)])
    
    return guide_down, guide_up


"""Command-line interface for crispio."""

from typing import (Callable, Dict, Tuple, Union)

from argparse import Namespace, FileType
from dataclasses import replace
from io import TextIOWrapper
import os
import sys

from bioino import FastaCollection, FastaSequence, GffFile
from carabiner import print_err
from carabiner.cliutils import CLIApp, CLICommand, CLIOption, clicommand
from tqdm.auto import tqdm
from streq import Circular

from .crosstalk import _get_mismatches
from .features import featurize
from .map import GuideLibrary
from .utils import sequences

__version__ = '0.0.2.post1'

def _allow_broken_pipe(f: Callable) -> Callable:

    def _f(*args, **kwargs):

        try:

            return f(*args, **kwargs)
        
        except BrokenPipeError:

            sys.exit(0)

    return _f


def _load_genome_and_gff(genome: TextIOWrapper, 
                         gff: TextIOWrapper) -> Tuple[FastaSequence, GffFile]:

    for fasta_sequence in FastaCollection.from_file(genome).sequences:
        break

    fasta_sequence = replace(fasta_sequence,
                             sequence=Circular(fasta_sequence.sequence))

    genome_filename = os.path.basename(genome.name)
    gff_filename = os.path.basename(gff.name)

    gff_data = GffFile.from_file(gff)

    new_metadata = [('genome-sequence', 'constrained', [fasta_sequence.name, 1, len(fasta_sequence.sequence)]), 
                    ('genome-description', 'constrained', [fasta_sequence.description]),
                    ('genome-filename', 'constrained', [genome_filename]),
                    ('sgRNA-map', 'free', [__package__, genome_filename, gff_filename])]
    old_metadata = list(gff_data.metadata.data)

    gff_data = replace(gff_data, 
                       lookup=True,
                       metadata=old_metadata + new_metadata)

    return fasta_sequence, gff_data


def _prepare_to_search(args: Namespace) -> Tuple[FastaSequence, GffFile, str, Dict[str, Union[str, int]]]:

    fasta_sequence, gff_data = _load_genome_and_gff(args.genome,
                                                    args.annotations)

    try:
        pam_search = sequences.pams[args.pam]
    except KeyError:
        pam_search = args.pam
    
    sgRNA_defaults = dict(seqid=fasta_sequence.name,
                          source=__package__,
                          feature='protospacer',
                          score='.',
                          phase='.')
    
    return fasta_sequence, gff_data, pam_search, sgRNA_defaults


@clicommand(message="Mapping sgRNAs with the following parameters")
def _map(args: Namespace) -> None:

    fasta_sequence, gff_data, pam_search, sgRNA_defaults = _prepare_to_search(args)

    guide_sequences = list(FastaCollection.from_file(args.input).sequences)
    n_guide_sequences = len(guide_sequences)

    print_err(f'\nFinding sgRNA sites matching {n_guide_sequences} sequences',
                f'from {args.input.name} and matching',
                f'PAM {args.pam} ({pam_search})',
                f'in {args.genome.name}...')

    gff_data.metadata.write(file=args.output)

    guide_library = GuideLibrary.from_mapping(guide_seq=guide_sequences,
                                              genome=fasta_sequence.sequence,
                                              pam_search=pam_search)

    for guide_match in guide_library.as_gff(max=1,
                                            annotations_from=gff_data,
                                            tags=args.attributes,
                                            gff_defaults=sgRNA_defaults):        
            
        _allow_broken_pipe(guide_match.write)(file=args.output)

    return None


@clicommand(message="Generating sgRNAs with the following parameters")
def _generate(args: Namespace) -> None:
    
    fasta_sequence, gff_data, pam_search, sgRNA_defaults = _prepare_to_search(args)

    gff_data.metadata.write(file=args.output)

    guide_library = GuideLibrary.from_generating(max_length=args.max_length,
                                                 min_length=args.min_length,
                                                 genome=fasta_sequence.sequence,
                                                 pam_search=pam_search)
    
    for guide_match in guide_library.as_gff(max=1,
                                            annotations_from=gff_data,
                                            tags=args.attributes,
                                            gff_defaults=sgRNA_defaults):        
            
        _allow_broken_pipe(guide_match.write)(file=args.output)
    
    return None


@clicommand(message="Featurizing sgRNAs with the following parameters")
def _featurize(args: Namespace) -> None:
    
    try:
        scaffold = sequences.scaffolds[args.scaffold]
    except KeyError:
        scaffold = args.scaffold

    print_err('> Generating features for guides...')

    input_gff = GffFile.from_file(args.input)
    input_gff.metadata.write(file=args.output)

    with tqdm(list(input_gff.lines)) as t:  ## run a progress bar

        for gff_line in t:

            t.set_postfix(current=gff_line.attributes["Name"][:40])
            
            new_features = featurize(gff_line, 
                                     scaffold=scaffold)
            attr = gff_line.attributes.copy()
            attr.update(new_features)
            
            gff_line = replace(gff_line, attributes=attr)

            _allow_broken_pipe(gff_line.write)(file=args.output)
    
    return None


@clicommand(message="Detecting off-targets with the following parameters")
def _offtarget(args: Namespace) -> None:

    gff1 = GffFile.from_file(args.input)
    gff2 = GffFile.from_file(args.gff2)

    gff2_lines = tuple(gff2.lines)
    
    n_mismatches = 0 
    pairs_checked = set()

    new_metadata = [('crosstalk-comparator', 'free', [os.path.basename(args.gff2.name)])]
    old_metadata = list(gff1.metadata.data)
    new_metadata = replace(gff1.metadata, 
                           data=old_metadata + new_metadata)
    new_metadata.write(file=args.output)

    with tqdm(gff1.lines) as t:
        
        for gff_line1 in t:

            mm = {}

            for gff_line2 in gff2_lines:

                t.set_postfix(mismatches=n_mismatches)

                pair, added_mm = _get_mismatches(gff_line1, gff_line2, 
                                                 maximum=args.mismatches,
                                                 pairs_checked=pairs_checked)

                pairs_checked.add(pair)

                mm.update(added_mm)
                n_mismatches = len(mm)

            attr = gff_line1.attributes.copy()
            
            if len(mm) > 0:
                attr['crosstalk'] = ('+'.join(f'{key}~{val}'for key, val in mm.items()))
            
            gff_line1 = replace(gff_line1, attributes=attr)

            _allow_broken_pipe(gff_line1.write)(args.output)
            
    return None


def main():

    length_max = CLIOption('--max_length', '-l', 
                           type=int, 
                           default=20,
                           help='Maximum length.')
    length_min = CLIOption('--min_length', '-m', 
                           type=int, 
                           default=None,
                           help='Minimum length.')
    
    # featurize
    scaffold = CLIOption('--scaffold', '-s', 
                        dest='scaffold', 
                        type=str, 
                        required=True,
                        help=('Name of a scaffold from "{}" or '
                              'a scaffold sequence.').format('", "'.join(sequences.scaffolds)))
    
    # offtarget
    gff2 = CLIOption('--gff2', '-2', 
                     type=FileType('r'),
                     required=True,
                     help='GFF file containing protospacers for comparison.')
    mismatches = CLIOption('--mismatches', '-e', 
                           type=int, 
                           default=2,
                           help='Maximum number of edits between guides to mark a close match.')

    inputs = CLIOption('input', 
                       type=FileType('r'),
                       default=sys.stdin,
                       nargs='?',
                       help='Input file. Default: STDIN.')

    pam = CLIOption('--pam', '-p',  
                    type=str, 
                    required=True,
                    help=('Either a PAM sequence or one of "{}".').format('", "'.join(sequences.pams)))
    genome = CLIOption('--genome', '-g', 
                       type=FileType('r'),
                       required=True,
                       help='Genome FASTA file')
    annotations = CLIOption('--annotations', '-a', 
                            type=FileType('r'),
                            required=True,
                            help='GFF file.')
    attributes = CLIOption('--attributes', '-t', 
                            type=str, 
                            nargs='*',
                            default=['Name', 'locus_tag', 'old_locus_tag', 'gene', 'gene_biotype'],
                            help='Tag to use in attribute field (column 9) of GFF file.')
    outputs = CLIOption('--output', '-o', 
                        type=FileType('w'),
                        default=sys.stdout,
                        help='Output file. Default: STDOUT')

    generate = CLICommand("generate", 
                          description="Generate and annotate all guide RNAs for a given genome.",
                          main=_generate,
                          options=[length_max, length_min, pam, genome, annotations, attributes, outputs])
    map_guides = CLICommand("map", 
                            description="Map and annotate provided guide RNAs to a given genome.",
                            main=_map,
                            options=[inputs, pam, genome, annotations, attributes, outputs])
    featurize = CLICommand("featurize", 
                            description="Annotate guide RNAs with additional calculated features.",
                            main=_featurize,
                            options=[inputs, scaffold, outputs])
    offtarget = CLICommand("offtarget", 
                           description="Compare two sets of guide RNAs for potential cross-target activity.",
                           main=_offtarget,
                           options=[inputs, gff2, mismatches, outputs])
    
    app = CLIApp('crispin', 
                 description='Design and analysis of bacterial CRISPRi experiments.',
                 version=__version__,
                 commands=[generate, map_guides, featurize, offtarget])

    app.run()

    return None


if __name__ == '__main__':

    main()


"""Tools for annotating guide RNAs from GFF data"""

from typing import Dict, Iterable, Mapping, Optional, Union

from dataclasses import asdict

from bioino import GffFile

Tguide = Dict[str, Union[str, int]]

_TAGS = ('Name', 'locus_tag', 'gene_biotype')
            
def annotate_from_gff(sgRNA: Tguide, 
                      gff_data: GffFile, 
                      tags: Optional[Iterable[str]] = None) -> Mapping:
    
    """Annotate dictionary of guide information with GFF annotations.

    Dictionary must at least have key 'pam_start' and 'pam_end' mapping to 
    numerical values.

    Parameters
    ----------
    sgRNA : dict
        Dictionary containing 'pam_start' and 'pam_end', and optionally other
        information about a guide.
    gff_data : bioino.GffFile
        GffFile object which was loaded with `lookup=True`.
    tags : list of str, optional
        Which GFF tags to extract from attributes of GFF features.

    Returns
    -------
    dict
        Guide RNA dictionary updated with GFF annotations.
    
    """
    
    tags = tags or _TAGS

    pam_loc = sgRNA['pam_start'] + abs(sgRNA['pam_start'] - sgRNA['pam_end']) // 2

    try:

        annotation_matches = gff_data._lookup[pam_loc][0]

    except IndexError:

        raise IndexError(f"Pam loc {pam_loc} is not annotated:\n"
                         f"{gff_data._lookup[pam_loc]}\n{gff_data._lookup[pam_loc - 1]}")

    for tag in tags:

        try:

            sgRNA[f'ann_{tag}'] = annotation_matches.attributes[tag]

        except KeyError:

            pass
        
    sgRNA['pam_offset'] = annotation_matches.attributes['offset']

    sgRNA.update({f'ann_{header}': val for header, val 
                  in asdict(annotation_matches.columns).items()})

    return sgRNA



"""Tools for detecting guide crosstalk."""

from typing import Iterable, Optional, Tuple

from streq import hamming

from bioino import GffLine

def _get_mismatches(gff1: GffLine, 
                    gff2: GffLine, 
                    pairs_checked: Optional[Iterable] = None,
                    maximum: int = 4,
                    seed_length: int = 4) -> Tuple[Tuple[str, str], int]:

    if pairs_checked is None:
        pairs_checked = set()

    query_id, query_seq, query_pam_start = (gff1.attributes['ID'], gff1.attributes['guide_sequence'], 
                                            gff1.attributes['pam_start'])
    ref_id, ref_seq, ref_pam_start = (gff2.attributes['ID'], gff2.attributes['guide_sequence'], 
                                      gff2.attributes['pam_start'])

    query_seed = query_seq[-seed_length:]
    mismatches = {}

    pair = tuple(sorted((query_id, ref_id)))

    if (query_pam_start != ref_pam_start and 
        pair not in pairs_checked and 
        query_id != ref_id): 

        same_seed = ref_seq.endswith(query_seed) 
        is_protospacer = gff2.columns.feature == 'protospacer'

        if is_protospacer and same_seed:

            distance = hamming(query_seq[:-seed_length], 
                               ref_seq[:-seed_length])
            
            close_match = distance <= maximum

            if close_match:

                mismatches = {ref_id: distance}

    return pair, mismatches



"""Utilities for crispio package."""

from collections import namedtuple
import os

import yaml

_data_path = os.path.join(os.path.dirname(__file__), 
                          'sequences.yml')

with open(_data_path, 'r') as f:
    _sequences = yaml.safe_load(f)

_sequence_dict = dict(
    pams = _sequences['PAMs'],
    scaffolds = _sequences['scaffolds']
)

SequenceCollection = namedtuple('SequenceCollection',
                                _sequence_dict)

sequences = SequenceCollection(**_sequence_dict)